use std::collections::HashMap;

use serde::{Deserialize, Serialize};

/// The body of the request for resources made to the rdsys backend
#[derive(Serialize)]
pub struct ResourceRequest {
    pub request_origin: String,
    pub resource_types: Vec<String>,
}

/// Representation of a bridge resource
#[derive(Deserialize, PartialEq, Eq, Debug)]
pub struct Resource {
    pub r#type: String,
    pub blocked_in: HashMap<String, bool>,
    pub protocol: String,
    pub address: String,
    pub port: u16,
    pub fingerprint: String,
    #[serde(rename = "or-addresses")]
    pub or_addresses: Option<Vec<String>>,
    pub distribution: String,
    pub flags: Option<HashMap<String, bool>>,
    pub params: Option<HashMap<String, String>>,
}

/// A ResourceDiff holds information about new, changed, or pruned resources
#[derive(Deserialize, PartialEq, Eq, Debug)]
pub struct ResourceDiff {
    pub new: Option<HashMap<String, Vec<Resource>>>,
    pub changed: Option<HashMap<String, Vec<Resource>>>,
    pub gone: Option<HashMap<String, Vec<Resource>>>,
    pub full_update: bool,
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn serialize_resource_request() {
        let req = ResourceRequest {
            request_origin: String::from("https"),
            resource_types: vec![String::from("obfs2"), String::from("scramblesuit")],
        };
        let json = serde_json::to_string(&req).unwrap();
        assert_eq!(
            json,
            "{\"request_origin\":\"https\",\"resource_types\":[\"obfs2\",\"scramblesuit\"]}"
        )
    }

    #[test]
    fn deserialize_resource() {
        let mut flags = HashMap::new();
        flags.insert(String::from("fast"), true);
        flags.insert(String::from("stable"), true);
        let mut params = HashMap::new();
        params.insert(
            String::from("password"),
            String::from("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"),
        );
        let bridge = Resource {
            r#type: String::from("scramblesuit"),
            blocked_in: HashMap::new(),
            protocol: String::from("tcp"),
            address: String::from("216.117.3.62"),
            port: 63174,
            fingerprint: String::from("BE84A97D02130470A1C77839954392BA979F7EE1"),
            or_addresses: None,
            distribution: String::from("https"),
            flags: Some(flags),
            params: Some(params),
        };

        let data = r#"
            {
                "type": "scramblesuit",
                "blocked_in": {},
                "protocol": "tcp",
                "address": "216.117.3.62",
                "port": 63174,
                "fingerprint": "BE84A97D02130470A1C77839954392BA979F7EE1",
                "or-addresses": null,
                "distribution": "https",
                "flags": {
                    "fast": true,
                    "stable": true
                },
                "params": {
                    "password": "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
                }
            }"#;
        let res: Resource = serde_json::from_str(data).unwrap();
        assert_eq!(bridge, res);
    }

    #[test]
    fn deserialize_resource_diff() {
        let data = r#"
            {
                "new": {
                    "obfs2": [
                        {
                            "type": "obfs2",
                            "blocked_in": {},
                            "Location": null,
                            "protocol": "tcp",
                            "address": "176.247.216.207",
                            "port": 42810,
                            "fingerprint": "10282810115283F99ADE5CFE42D49644F45D715D",
                            "or-addresses": null,
                            "distribution": "https",
                            "flags": {
                                "fast": true,
                                "stable": true,
                                "running": true,
                                "valid": true
                            }
                        },
                        {
                            "type": "obfs2",
                            "blocked_in": {},
                            "protocol": "tcp",
                            "address": "133.69.16.145",
                            "port": 58314,
                            "fingerprint": "BE84A97D02130470A1C77839954392BA979F7EE1",
                            "or-addresses": null,
                            "distribution": "https",
                            "flags": {
                                "fast": true,
                                "stable": true,
                                "running": true,
                                "valid": true
                            }
                        }
                    ],
                    "scramblesuit": [
                        {
                            "type": "scramblesuit",
                            "blocked_in": {},
                            "protocol": "tcp",
                            "address": "216.117.3.62",
                            "port": 63174,
                            "fingerprint": "BE84A97D02130470A1C77839954392BA979F7EE1",
                            "or-addresses": null,
                            "distribution": "https",
                            "flags": {
                                "fast": true,
                                "stable": true,
                                "running": true,
                                "valid": true
                            },
                            "params": {
                                "password": "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567"
                            }
                        }
                    ]
                },
                "changed": null,
                "gone": null,
                "full_update": true
            }"#;
        let diff: ResourceDiff = serde_json::from_str(data).unwrap();
        assert_ne!(diff.new, None);
        assert_eq!(diff.changed, None);
        assert_eq!(diff.gone, None);
        assert_eq!(diff.full_update, true);
        if let Some(new) = diff.new {
            assert_eq!(new["obfs2"][0].r#type, "obfs2");
        }
    }

    #[test]
    fn deserialize_empty_resource_diff() {
        let data = r#"
            {
                "new": null,
                "changed": null,
                "gone": null,
                "full_update": true
            }"#;
        let diff: ResourceDiff = serde_json::from_str(data).unwrap();
        let empty_diff = ResourceDiff {
            new: None,
            changed: None,
            gone: None,
            full_update: true,
        };
        assert_eq!(diff, empty_diff);
    }

    #[test]
    fn deserialize_empty_condensed_diff() {
        let data = "{\"new\": null,\"changed\": null,\"gone\": null,\"full_update\": true}";
        let diff: ResourceDiff = serde_json::from_str(data).unwrap();
        let empty_diff = ResourceDiff {
            new: None,
            changed: None,
            gone: None,
            full_update: true,
        };
        assert_eq!(diff, empty_diff);
    }
}
